@@ -9,43 +9,38 @@ import logging
 logger = logging.getLogger("proxmin")
 
 
-def _copy_tuple(X):
-    return tuple(item.copy() for item in X)
-
-
-def _as_tuple(X):
-    if type(X) in [list, tuple]:
-        return X
-    else:
-        return (X,)
-
-
 def pgm(
     X,
     grad,
     step,
     prox=None,
     accelerated=False,
-    relax=None,
+    backtracking=False,
+    f=None,
     e_rel=1e-6,
     max_iter=1000,
     callback=None,
 ):
     """Proximal Gradient Method
 
-    Adapted from Combettes 2009, Algorithm 3.4.
-    The accelerated version is Algorithm 3.6 with modifications
-    from Xu & Yin (2015).
+    Adapted from Parikh & Boyd (2014, section 4.2)
+    The accelerated version is FISTA (Beck & Teboulle, 2009)
+    with block coordinate descent from from Xu & Yin (2015).
 
     Args:
         X: initial X, will be updated
         grad: gradient function of f wrt to X
+            Signature: grad(*X) -> df/dX
         step: function to compute step size.
             Should be smaller than 2/L with L the Lipschitz constant of grad
-            Signature: step(*X, it=None) -> float
+            Signature: step(*X, it=None, grad=None) -> float
         prox: proximal operator for penalty functions
+            If X is a list for block coordinate PGM, prox can be a list as well.
+            Signature: prox(X, step) -> X'
         accelerated: if Nesterov acceleration should be used
-        relax: (over)relaxation parameter, 0 < relax < 1.5
+        backtracking: if backtracking should be used to determine step sizes
+        f: smooth function, only needed for backtracking
+            Signature: f(*X) -> float
         e_rel: relative error of X sufficient for convergence
         max_iter: maximum iteration
         callback: arbitrary logging function
@@ -57,26 +52,37 @@ def pgm(
         step: last iteration steps
     """
     # Set up: turn X and prox into tuples
-    X = _as_tuple(X)
+    X = utils._as_tuple(X)
     N = len(X)
-    prox = _as_tuple(prox)
+    prox = utils._as_tuple(prox)
     if len(prox) == 1:
         prox = prox * N
     assert len(prox) == len(X)
 
+    # convert prox=None to prox_id, so that we don't have to test later on
+    from .operators import prox_id
+    prox = tuple(p if p is not None else prox_id for p in prox )
+
     if np.isscalar(e_rel):
         e_rel = (e_rel,) * N
-
     assert len(e_rel) == len(X)
 
-    if relax is not None:
-        assert relax > 0 and relax < 1.5
+    assert backtracking is False or f is not None
+
+    # check if step function has grads argument
+    try:
+        step(*X, it=0, grads=X)
+        _step = step
+    except TypeError:
+        _step = lambda *X, it=None, grads=None: step(*X, it=it)
 
     if callback is None:
         callback = utils.NullCallback()
 
     # init
-    stepper = utils.NesterovStepper(accelerated=accelerated)
+    accel = utils.NesterovAccelerator(accelerated=accelerated)
+    # for backtracking
+    T = [1.,] * N
 
     for it in range(max_iter):
 
@@ -84,27 +90,41 @@ def pgm(
             callback(*X, it=it)
 
             # use Nesterov acceleration (if omega > 0), automatically incremented
-            omega = stepper.omega
+            omega = accel.omega
             if omega > 0:
                 _X = tuple(X[j] + omega * (X[j] - X_[j]) for j in range(N))
+            elif backtracking:
+                _X = utils._copy_tuple(X)
             else:
                 _X = X
 
             # make copy for convergence test and acceleration
-            X_ = _copy_tuple(X)
-
-            # (P)GM step
-            G = _as_tuple(grad(*_X))
-            S = _as_tuple(step(*_X, it=it))
+            X_ = utils._copy_tuple(X)
 
+            # PGM step
+            G = utils._as_tuple(grad(*_X))
+            S = utils._as_tuple(_step(*_X, it=it, grads=G))
             for j in range(N):
-                _X[j][:] -= S[j] * G[j]
+                X[j][:] = prox[j](_X[j] - T[j]*S[j] * G[j], T[j]*S[j])
 
-                if prox[j] is not None:
-                    X[j][:] = prox[j](_X[j], S[j])
+            if backtracking:
+                # Beck & Teboulle, eq. 3.2
+                f_now = f(*X)
+                if it == 0:
+                    f_prev = f(*X_)
 
-                if relax is not None:
-                    X[j][:] += (relax - 1) * (X[j] - X_[j])
+                # dropping g from F and Q because it cancels
+                while  f_now > f_prev + \
+                    np.sum([ np.sum((X[j] - X_[j])*G[j]) + 0.5 / (T[j]*S[j]) * np.sum((X[j] - X_[j])**2)  for j in range(N) ]):
+
+                    # find largest relative update direction
+                    jmax = np.argmax([ np.max(np.abs(S[j]*G[j]))/np.max(np.abs(X_[j])) for j in range(N) ])
+                    T[jmax] /= 2
+                    X[jmax][:] = prox[jmax](_X[jmax] - T[jmax]*S[jmax] * G[jmax], T[jmax]*S[jmax])
+                    f_now = f(*X)
+
+                # save f_prev from current function evaluation
+                f_prev = f_now
 
             # test for fixed point convergence
             converged = tuple(
@@ -135,6 +155,17 @@ def _adam_phi_psi(it, G, M, V, Vhat, b1, b2, eps, p):
     Psi = np.sqrt(V / (1 - b2 ** t)) + eps
     return Phi, Psi
 
+def _nadam_phi_psi(it, G, M, V, Vhat, b1, b2, eps, p):
+    # moving averages
+    M[:] = (1 - b1[it]) * G + b1[it] * M
+    V[:] = (1 - b2) * (G ** 2) + b2 * V
+
+    # bias correction
+    t = it + 1
+    Phi = (b1[it] * M[:] + (1- b1[it]) * G) / (1 - b1[it] ** t)
+    Psi = np.sqrt(V / (1 - b2 ** t)) + eps
+    return Phi, Psi
+
 
 def _amsgrad_phi_psi(it, G, M, V, Vhat, b1, b2, eps, p):
     # moving averages
@@ -238,6 +269,7 @@ def adaprox(
     Uses multiple variants of adaptive quasi-Newton gradient descent
 
         * Adam (Kingma & Ba 2015)
+        * NAdam (Dozat 2016)
         * AMSGrad (Reddi, Kale & Kumar 2018)
         * PAdam (Chen & Gu 2018)
         * AdamX (Phuong & Phong 2019)
@@ -257,7 +289,7 @@ def adaprox(
             Should be smaller than 2/L with L the Lipschitz constant of grad
             Signature: step(*X, it=None) -> float
         prox: proximal operator of penalty function
-        scheme: one of ["adam", "adamx", "amsgrad", "padam","radam"]
+        scheme: one of ["adam", "nadam", "adamx", "amsgrad", "padam", "radam"]
         b1: (float or array) first moment momentum decay
         b2: second moment momentum decay
         eps: softening of second moment (only for algorithm == "adam")
@@ -277,10 +309,13 @@ def adaprox(
         M: last iteration gradient
         V: last iteration squared gradient
         Vhat: last iteration maximized squared gradient
+
+    Reference:
+        Melchior, Joseph & Moolekamp, Algorithm 1 (arXiv:1910.10094)
     """
-    X = _as_tuple(X)
+    X = utils._as_tuple(X)
     N = len(X)
-    prox = _as_tuple(prox)
+    prox = utils._as_tuple(prox)
     if len(prox) == 1:
         prox = prox * N
     assert len(prox) == len(X)
@@ -298,10 +333,11 @@ def adaprox(
     assert eps >= 0
     assert p > 0 and p <= 0.5
     scheme = scheme.lower()
-    assert scheme in ["adam", "adamx", "amsgrad", "padam", "radam"]
+    assert scheme in ["adam", "nadam", "adamx", "amsgrad", "padam", "radam"]
 
     phi_psi = {
         "adam": _adam_phi_psi,
+        "nadam": _nadam_phi_psi,
         "amsgrad": _amsgrad_phi_psi,
         "padam": _padam_phi_psi,
         "adamx": _adamx_phi_psi,
@@ -330,10 +366,10 @@ def adaprox(
 
         try:
             callback(*X, it=it)
-            G = _as_tuple(grad(*X))
-            Alpha = _as_tuple(step(*X, it=it))
+            G = utils._as_tuple(grad(*X))
+            Alpha = utils._as_tuple(step(*X, it=it))
             if check_convergence:
-                X_ = _copy_tuple(X)
+                X_ = utils._copy_tuple(X)
 
             for j in range(N):
                 Phi, Psi = phi_psi[scheme](
